/*
 *  This code was written by Yufei Yuan <yuanyufei@hotmail.com>
 */

#include "lcmencode.h"

#define USG_STRING "[-w %s:wavelet] [-b %s:boundary] [-pw %: %s:pcpfile] [-nl %d:num_levels] [-m %: %s:mask] [-vo %:] %f:rate %s:imgfile %s:bitstream"

QccWAVWavelet Wavelet;
QccString WaveletFilename = QCCWAVWAVELET_DEFAULT_WAVELET;
QccString Boundary = "symmetric";

QccWAVPerceptualWeights PerceptualWeights;
int UsePerceptualWeights = 0;

int NumLevels = 5;
int NumSubbands;

QccIMGImage InputImage;
int ImageNumRows, ImageNumCols;

QccIMGImage Mask;
int MaskSpecified = 0;
int MaskNumRows, MaskNumCols;

QccBitBuffer OutputBuffer;

int ValueOnly = 0;

float TargetRate;
float ActualRate;
int NumPixels;
int TargetBitCnt;

int main(int argc, char *argv[])
{
	int row, col;

	QccInit(argc, argv);

	QccWAVlcmHeader();

	QccWAVWaveletInitialize(&Wavelet);
	QccWAVPerceptualWeightsInitialize(&PerceptualWeights);
	QccIMGImageInitialize(&InputImage);
	QccIMGImageInitialize(&Mask);
	QccBitBufferInitialize(&OutputBuffer);

	if(QccParseParameters(argc, argv, USG_STRING, WaveletFilename, Boundary, 
		&UsePerceptualWeights, PerceptualWeights.filename, &NumLevels, &MaskSpecified,
		Mask.filename, &ValueOnly, &TargetRate, InputImage.filename, OutputBuffer.filename))
		QccErrorExit();

	if(NumLevels < 0) {
		QccErrorAddMessage("%s: Number of levels of decomposition must be nonnegative",
			argv[0]);
		QccErrorExit();
	}

	NumSubbands = QccWAVSubbandPyramidNumLevelsToNumSubbands(NumLevels);

	if(QccWAVWaveletCreate(&Wavelet, WaveletFilename, Boundary)) {
		QccErrorAddMessage("%s: Error calling QccWAVWaveletCreate()", argv[0]);
		QccErrorExit();
	}

	if(UsePerceptualWeights) {
		if(QccWAVPerceptualWeightsRead(&PerceptualWeights)) {
			QccErrorAddMessage("%s: Error calling QccWAVPerceptualWeightsRead()",
				argv[0]);
			QccErrorExit();
		}
		if(PerceptualWeights.num_subbands != NumSubbands) {
			QccErrorAddMessage("%s: Number of subbands in %s does not match specified number of levels (%d)",
				argv[0], PerceptualWeights.filename, NumLevels);
		}
	}

	if(QccIMGImageRead(&InputImage)) {
		QccErrorAddMessage("%s: Error calling QccIMGImageRead()", argv[0]);
		QccErrorExit();
	}
	if(QccIMGImageColor(&InputImage)) {
		QccErrorAddMessage("%s: Color images not supported", argv[0]);
		QccErrorExit();
	}
	if(QccIMGImageGetResolution(&InputImage, &ImageNumRows, &ImageNumCols)) {
		QccErrorAddMessage("%s: Error calling QccIMGImageGetResolution()", argv[0]);
		QccErrorExit();
	}
	if(ImageNumCols != ((int)(ImageNumCols >> NumLevels) << NumLevels)) {
		QccErrorAddMessage("%s: Image resolution (%d x %d) is not a multiple of %d (as needed for a decomposition of %d levels)",
			argv[0], ImageNumCols, ImageNumRows, (1 << NumLevels), NumLevels);
		QccErrorExit();
	}
	if(ImageNumRows != ((int)(ImageNumRows >> NumLevels) << NumLevels)) {
		QccErrorAddMessage("%s: Image resolution (%d x %d) is not a multiple of %d (as needed for a decomposition of %d levels",
			argv[0], ImageNumCols, ImageNumRows, (1 << NumLevels), NumLevels);
		QccErrorExit();
	}

	if(MaskSpecified) {
		if(QccIMGImageRead(&Mask)) {
			QccErrorAddMessage("%s: Error calling QccIMGImageRead()", argv[0]);
			QccErrorExit();
		}
		if(QccIMGImageColor(&Mask)) {
			QccErrorAddMessage("%s: Mask must be grayscale", argv[0]);
			QccErrorExit();
		}
		if(QccIMGImageGetResolution(&Mask, &MaskNumRows, &MaskNumCols)) {
			QccErrorAddMessage("%s: Error calling QccIMGImageGetResolution()", argv[0]);
			QccErrorExit();
		}

		if((MaskNumRows != ImageNumRows) || (MaskNumCols != ImageNumCols)) {
			QccErrorAddMessage("%s: Mask must be same size as image", argv[0]);
			QccErrorExit();
		}

		NumPixels = 0;

		for (row = 0; row < MaskNumRows; row++) {
			for (col = 0; col < MaskNumCols; col++) {
				if(!QccAlphaTransparent(Mask.Y.image[row][col])) NumPixels++;
			}
		}
	}
	else NumPixels = ImageNumRows * ImageNumCols;

	OutputBuffer.type = QCCBITBUFFER_OUTPUT;
	if(QccBitBufferStart(&OutputBuffer)) {
		QccErrorAddMessage("%s: Error calling QccBitBufferStart()", argv[0]);
		QccErrorExit();
	}

	TargetBitCnt = (int)(ceil(((double)NumPixels * TargetRate)/8.0)) * 8;
	/* fprintf(stderr, "The target bit count is: %d\n", TargetBitCnt); */
	/* TargetBitCnt = (int)(NumPixels * TargetRate / 8.0) * 8; */

	if(QccWAVlcmEncode(&(InputImage.Y), (MaskSpecified ? &(Mask.Y) : NULL), &OutputBuffer,
		NumLevels, &Wavelet, ((UsePerceptualWeights) ? &PerceptualWeights : NULL), 
		TargetBitCnt)) {
		QccErrorAddMessage("%s: Error calling QccWAVlcmEncode()", argv[0]);
		QccErrorExit();
	}

	ActualRate = (double)OutputBuffer.bit_cnt / ImageNumRows / ImageNumCols;

	if(QccBitBufferEnd(&OutputBuffer)) {
		QccErrorAddMessage("%s: Error calling QccBitBufferEnd()", argv[0]);
		QccErrorExit();
	}

	if(ValueOnly) printf("%f\n", ActualRate);
	else {
		printf("LCM coding of %s:\n", InputImage.filename);
		printf("  Target rate: %f bpp\n", TargetRate);
		printf("  Actual rate: %f bpp\n", ActualRate);
	}

	QccExit;
}
